/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CaQaBuilder.h
/// \date   20.11.2023
/// \brief  A QA module for CA tracking (header)
/// \author S.Zharko <s.zharko@gsi.de>

#pragma once

#include "CaHit.h"  // for HitIndex_t
#include "CaQaBuilder.h"
#include "CaQaData.h"
#include "CaVector.h"

namespace cbm::algo::ca
{
  class Parameters;
  class InputData;
  class Track;
}  // namespace cbm::algo::ca

namespace cbm::algo::ca::qa
{
  /// \class cbm::algo::ca::qa::Builder
  /// \brief Builder class for the CA tracking QA (header)
  ///
  class Builder {
  public:
    using TrackV_t    = ca::Vector<ca::Track>;
    using HitIndexV_t = ca::Vector<std::vector<std::pair<uint32_t, uint32_t>>>;

    /// \brief Default destructor
    Builder() = delete;

    /// \brief Constructor from the configuration object
    /// \param config  QA configuration object
    Builder(const Config& config) : fConfig(config) {}

    /// \brief Copy constructor
    Builder(const Builder&) = delete;

    /// \brief Move constructor
    Builder(Builder&&) = delete;

    /// \brief Destructor
    ~Builder() = default;

    /// \brief Copy assignment operator
    Builder& operator=(const Builder&) = delete;

    /// \brief Move assignment operator
    Builder& operator=(Builder&&) = delete;

    /// \brief QA execution function
    Data Build();

    /// \brief Gets QA config
    const Config& GetConfig() const { return fConfig; }

    /// \brief Registers tracking input data
    /// \note  Call per TS
    void RegisterInputData(const InputData* pInputData) { fpInputData = pInputData; }

    /// \brief Registers track vector
    /// \note  Call per TS
    void RegisterTracks(const Vector<Track>* pvTracks) { fpvTracks = pvTracks; }

    /// \brief Registers reco hits indices
    /// \note  Call per TS
    void RegisterRecoHitIndices(const Vector<HitIndex_t>* pvRecoHits) { fpvRecoHits = pvRecoHits; }

    /// \brief Registers tracking parameters object
    /// \note  Call per run
    void RegisterParameters(const Parameters* pParameters) { fpParameters = pParameters; }

  private:
    Config fConfig;                                   ///< QA configuration
    const Parameters* fpParameters        = nullptr;  ///< Pointer to tracking parameters
    const Vector<Track>* fpvTracks        = nullptr;  ///< Pointer to tracks vector
    const InputData* fpInputData          = nullptr;  ///< Pointer to input data
    const Vector<HitIndex_t>* fpvRecoHits = nullptr;  ///< Pointer to reco hit indices
  };
}  // namespace cbm::algo::ca::qa
