/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CaQaDefinitions.h
/// \date   20.11.2023
/// \brief  Definitions (enums and structures) for the cbm::algo::ca::qa namespace
/// \author S.Zharko <s.zharko@gsi.de>

#pragma once

#include <iomanip>
#include <sstream>

#include "CaEnumArray.h"

namespace cbm::algo::ca::qa
{
  /// \struct H1Pars
  /// \brief  Parameters of the 1D-histograms
  ///
  struct H1Pars {
    std::string fName;   ///< Histogram name
    std::string fTitle;  ///< Histogram title
    uint32_t fNumBins;   ///< Number of bins in histogram
    double fMinValue;    ///< Minimal value
    double fMaxValue;    ///< Maximal value

    /// \brief String representation of the object
    std::string ToString() const
    {
      using std::setw;
      std::stringstream msg;
      msg << "nbins " << setw(8) << fNumBins << ", min " << setw(15) << fMinValue << ", max " << setw(15) << fMaxValue;
      return msg.str();
    };
  };

  /// \enum  H1Key
  /// \brief Keys for the 1D-histograms
  ///
  enum class H1Key
  {
    //TrackFstTx,     ///< First hit: Slope of the track along x-axis
    //TrackFstTy,     ///< First hit: Slope of the track along y-axis
    FstTrkTheta,  ///< First hit: Polar angle of the track
    FstTrkPhi,    ///< First hit: Azimuthal angle of the track
    FstChi2Ndf,   ///< First hit: chi2 / NDF
    FstHitSta,    ///< Station index of first hit
    LstHitSta,    ///< Station index of last hit
    TrkNofHits,   ///< Number of hits in track
    kEND          ///< END
  };

  /// \brief  Alias to array, indexed by H1Key
  template<typename T>
  using H1KeyArray_t = EnumArray<H1Key, T>;

}  // namespace cbm::algo::ca::qa
