/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Volker Friese, Dominik Smith [committer], Sebastian Heinemann, Felix Weiglhofer */


#ifndef UNPACK_H
#define UNPACK_H 1

#include "tof/ReadoutConfig.h"
#include "tof/Unpack.h"
#include "trd/ReadoutConfig.h"
#include "trd/Unpack.h"
#include "trd2d/ReadoutConfig.h"
#include "trd2d/Unpack.h"

#include <gsl/span>
#include <optional>
#include <sstream>
#include <vector>

#include <xpu/host.h>

#include "DigiData.h"
#include "PODVector.h"
#include "bmon/ReadoutConfig.h"
#include "bmon/Unpack.h"
#include "much/ReadoutConfig.h"
#include "much/Unpack.h"
#include "rich/ReadoutConfig.h"
#include "rich/Unpack.h"
#include "sts/Digi.h"
#include "sts/ReadoutConfigLegacy.h"
#include "sts/Unpack.h"

namespace cbm::algo
{

  /** @struct UnpackMonitorData
   ** @author Dominik Smith <d.smith@gsi.de>
   ** @since 2 Jun 2023
   ** @brief Monitoring data for unpacking
   **/
  struct UnpackMonitorData {
    std::vector<sts::UnpackMonitorData> fSts;      ///< Monitoring data for STS
    std::vector<much::UnpackMonitorData> fMuch;    ///< Monitoring data for MUCH
    std::vector<tof::UnpackMonitorData> fTof;      ///< Monitoring data for TOF
    std::vector<bmon::UnpackMonitorData> fBmon;    ///< Monitoring data for Bmon
    std::vector<trd::UnpackMonitorData> fTrd;      ///< Monitoring data for TRD
    std::vector<trd2d::UnpackMonitorData> fTrd2d;  ///< Monitoring data for TRD2D
    std::vector<rich::UnpackMonitorData> fRich;    ///< Monitoring data for RICH
    xpu::timings fTime;
    size_t fNumMs       = 0;
    size_t fNumBytes    = 0;
    size_t fNumBytesInSts       = 0;
    size_t fNumBytesInMuch      = 0;
    size_t fNumBytesInTof       = 0;
    size_t fNumBytesInBmon      = 0;
    size_t fNumBytesInTrd       = 0;
    size_t fNumBytesInTrd2d     = 0;
    size_t fNumBytesInRich      = 0;
    size_t fNumDigis    = 0;
    size_t fNumCompUsed = 0;
    size_t fNumErrInvalidEqId   = 0;
    size_t fNumErrInvalidSysVer = 0;
    std::string print() const
    {
      std::stringstream ss;
      ss << "TS stats: num MS " << fNumMs << ", bytes " << fNumBytes << ", digis " << fNumDigis << ", components "
         << fNumCompUsed << ", invalidEqIds " << fNumErrInvalidEqId << ", invalidSysVersions " << fNumErrInvalidSysVer
         << std::endl;
      return ss.str();
    }
  };

  /** @class Unpack
 ** @brief Algo class for unpacking digi timeslicess
 ** @author Dominik Smith <d.smith@gsi.de>
 ** @since 02.06.2023
 **
 **/
  class Unpack {

  public:
    typedef std::pair<DigiData, UnpackMonitorData> resultType;
    using Subsystem = fles::Subsystem;

    /** @brief Algorithm execution
     ** @param fles timeslice to unpack
     ** @return pair: digi timeslice, monitoring data
     **/
    resultType operator()(const fles::Timeslice* timeslice);

    /** @brief Default constructor **/
    Unpack() {};

    /** @brief Destructor **/
    ~Unpack() {};

    /** @brief Parameters for STS unpackers **/
    sts::ReadoutConfigLegacy fStsConfig {};

    /** @brief Parameters for MUCH unpackers **/
    much::ReadoutConfig fMuchConfig {};

    /** @brief Parameters for TOF unpackers **/
    tof::ReadoutConfig fTofConfig {};

    /** @brief Parameters for Bmon unpackers **/
    bmon::ReadoutConfig fBmonConfig {};

    /** @brief Parameters for TRD unpackers **/
    trd::ReadoutConfig fTrdConfig {};

    /** @brief Parameters for TRD2D unpackers **/
    trd2d::ReadoutConfig fTrd2dConfig {};

    /** @brief Parameters for RICH unpackers **/
    rich::ReadoutConfig fRichConfig {};


    /**
     * @brief Set whether to apply walk correction. Must be set before Init(). (default: true)
    **/
    void SetApplyWalkCorrection(bool applyWalkCorrection) { fApplyWalkCorrection = applyWalkCorrection; }

    /** @brief Initialize unpackers and fill parameters from config objects
     * @param subIds: vector of subsystem identifiers to unpack, default: all
     * @see Init()
     **/
    void Init(std::vector<fles::Subsystem> subIds = {
                Subsystem::STS,
                Subsystem::MUCH,
                Subsystem::TOF,
                Subsystem::BMON,
                Subsystem::TRD,
                Subsystem::TRD2D,
                Subsystem::RICH,
              });

    bool DetectorEnabled(Subsystem subsystem)
    {
      return std::find(fSubsystems.begin(), fSubsystems.end(), subsystem) != fSubsystems.end();
    }

  private:  // methods
    /** @brief Microslice loop **/
    template<class Digi, class UnpackAlgo, class MonitorData>
    size_t MsLoop(const fles::Timeslice* timeslice, std::map<uint16_t, UnpackAlgo>& algoMap, const uint64_t comp,
                  const uint16_t eqId, PODVector<Digi>* digis, UnpackMonitorData& monitor,
                  std::vector<MonitorData>* monitorMs, uint8_t sys_ver);

    /** @brief Parallel microslice loop **/
    template<class Digi, class UnpackAlgo, class Monitor>
    size_t ParallelMsLoop(const Subsystem subsystem, UnpackMonitorData& monitor, PODVector<Digi>& digisOut,
                          std::vector<Monitor>& monitorOut, const fles::Timeslice& ts,
                          const std::map<u16, UnpackAlgo>& algos, u8 sys_ver);

    std::pair<size_t, size_t> ParallelInit(const fles::Timeslice& ts, Subsystem subsystem,
                                           gsl::span<const uint16_t> legalEqIds, uint8_t sys_ver,
                                           UnpackMonitorData& monitor, std::vector<u16>& eqIds,
                                           std::vector<fles::MicrosliceDescriptor>& msDesc,
                                           std::vector<const u8*>& msContent);

    /** @brief Sort Digis and add bytes to timer for throughput */
    template<class Digi>
    void DoSort(PODVector<Digi>& digis);


  private:                                                  // members
    bool fApplyWalkCorrection                      = true;  ///< Apply walk correction
    std::vector<Subsystem> fSubsystems             = {};    ///< Detector identifiers to unpack

    /** @brief STS unpackers **/
    std::map<uint16_t, sts::Unpack> fAlgoSts = {};

    /** @brief MUCH unpackers **/
    std::map<uint16_t, much::Unpack> fAlgoMuch = {};

    /** @brief TOF unpackers **/
    std::map<uint16_t, tof::Unpack> fAlgoTof = {};

    /** @brief Bmon unpackers **/
    std::map<uint16_t, bmon::Unpack> fAlgoBmon = {};

    /** @brief TRD unpackers **/
    std::map<uint16_t, trd::Unpack> fAlgoTrd = {};

    /** @brief TRD2D unpackers **/
    std::map<uint16_t, trd2d::Unpack> fAlgoTrd2d = {};

    /** @brief RICH unpackers **/
    std::map<uint16_t, rich::Unpack> fAlgoRich = {};

    /** @brief System time offsets **/
    std::map<Subsystem, int32_t> fSystemTimeOffset = {
      {Subsystem::STS, -970}, {Subsystem::MUCH, -980}, {Subsystem::RICH, 100},  {Subsystem::TOF, 40},
      {Subsystem::BMON, 0},   {Subsystem::TRD, 1300},  {Subsystem::TRD2D, -510}};

  private:  // methods
    template<typename UnpackAlgo>
    std::vector<uint16_t> GetEqIds(const std::map<uint16_t, UnpackAlgo>& algoMap)
    {
      std::vector<uint16_t> eqIds;
      eqIds.reserve(algoMap.size());
      for (const auto& [eqId, algo] : algoMap) {
        eqIds.push_back(eqId);
      }
      return eqIds;
    }
  };
}  // namespace cbm::algo

#endif /* UNPACK_H */
