/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CbmCaInputQaMuch.cxx
/// \date   13.01.2023
/// \brief  QA-task for CA tracking input from MuCh detector (implementation)
/// \author S.Zharko <s.zharko@gsi.de>

#include "CbmCaInputQaMuch.h"
#include "CbmMuchTrackingInterface.h"

ClassImp(CbmCaInputQaMuch);

// ---------------------------------------------------------------------------------------------------------------------
//
CbmCaInputQaMuch::CbmCaInputQaMuch(int verbose, bool isMCUsed) : CbmCaInputQaBase("CbmCaInputQaMuch", verbose, isMCUsed)
{
  // Default parameters of task
  DefineParameters();
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmCaInputQaMuch::DeInit() { CbmCaInputQaBase::DeInit(); }

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmCaInputQaMuch::DefineParameters()
{
  auto SetRange = [](std::array<double, 2>& range, double min, double max) {
    range[0] = min;
    range[1] = max;
  };
  // Hit errors
  SetRange(fRHitDx, 0.0000, 5.00);  // [cm]
  SetRange(fRHitDy, 0.0000, 5.00);  // [cm]
  SetRange(fRHitDu, 0.0000, 5.00);  // [cm]
  SetRange(fRHitDv, 0.0000, 5.00);  // [cm]
  SetRange(fRHitDt, 0.0000, 0.15);  // [ns]
  // Residuals
  SetRange(fRResX, -2.00, 2.00);
  SetRange(fRResY, -4.00, 4.00);
  SetRange(fRResU, -2.00, 2.00);
  SetRange(fRResV, -4.00, 4.00);
  SetRange(fRResT, -5.00, 5.00);
  SetRange(fRangeDzHitPoint, -50., 50.);
  // QA result selection criteria
  SetRange(fEffRange, 10.0, 30.0);  ///< Range for hit efficiency approximation
  fResMeanThrsh   = 0.50;           ///< Maximum allowed deviation of residual mean from zero [sigma]
  fPullMeanThrsh  = 0.10;           ///< Maximum allowed deviation of pull mean from zero
  fPullWidthThrsh = 2.00;           ///< Maximum allowed deviation of pull width from unity
  fEffThrsh       = 0.50;           ///< Threshold for hit efficiency in the selected range
  fMaxDiffZStHit  = 1.00;           ///< Maximum allowed difference between z-position of hit and station [cm]
  fMinMomentum    = 0.05;           ///< Minimum momentum of particle [GeV/c]
  // Track selection criteria
  fTrackSelectionPrimary  = true;  ///< must the track come from the primary vertex
  fTrackSelectionMomentum = 0.1;   ///< track momentum GeV when it exits the tracking station
  fTrackSelectionAngle    = 60.;   ///< track incl. angle [grad] when it exits the station
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmCaInputQaMuch::FillHistogramsPerHit() {}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmCaInputQaMuch::InitDataBranches()
{
  // STS tracking detector interface
  fpDetInterface = CbmMuchTrackingInterface::Instance();

  auto baseInitStatus = CbmCaInputQaBase::InitDataBranches();
  if (kSUCCESS != baseInitStatus) { return baseInitStatus; }

  return kSUCCESS;
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmCaInputQaMuch::InitCanvases()
{
  auto baseInitStatus = CbmCaInputQaBase::InitCanvases();
  if (kSUCCESS != baseInitStatus) { return baseInitStatus; }

  return kSUCCESS;
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmCaInputQaMuch::InitHistograms()
{
  auto baseInitStatus = CbmCaInputQaBase::InitHistograms();
  if (kSUCCESS != baseInitStatus) { return baseInitStatus; }

  return kSUCCESS;
}
