/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// @file   CbmCaInputQaTrd.cxx
/// @date   13.01.2023
/// @brief  QA-task for CA tracking input from TRD detector (implementation)
/// @author S.Zharko <s.zharko@gsi.de>

#include "CbmCaInputQaTrd.h"
#include "CbmTrdTrackingInterface.h"

ClassImp(CbmCaInputQaTrd);

// ---------------------------------------------------------------------------------------------------------------------
//
CbmCaInputQaTrd::CbmCaInputQaTrd(int verbose, bool isMCUsed) : CbmCaInputQaBase("CbmCaInputQaTrd", verbose, isMCUsed)
{
  // Default parameters of task
  DefineParameters();
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmCaInputQaTrd::DeInit() { CbmCaInputQaBase::DeInit(); }

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmCaInputQaTrd::DefineParameters()
{
  auto SetRange = [](std::array<double, 2>& range, double min, double max) {
    range[0] = min;
    range[1] = max;
  };
  // Hit errors
  SetRange(fRHitDx, 0.0000, 5.00);   // [cm]
  SetRange(fRHitDy, 0.0000, 5.00);   // [cm]
  SetRange(fRHitDu, 0.0000, 5.00);   // [cm]
  SetRange(fRHitDv, 0.0000, 5.00);   // [cm]
  SetRange(fRHitDt, 0.0000, 100.00);  // [ns]
  // Residuals
  SetRange(fRResX, -10.00, 10.00);
  SetRange(fRResY, -10.00, 10.00);
  SetRange(fRResU, -2.00, 2.00);
  SetRange(fRResV, -10.00, 10.00);
  SetRange(fRResT, -200.0, 200.0);
  // QA result selection criteria
  SetRange(fEffRange, 10.0, 30.0);  ///< Range for hit efficiency approximation
  fResMeanThrsh   = 0.50;           ///< Maximum allowed deviation of residual mean from zero [sigma]
  fPullMeanThrsh  = 0.10;           ///< Maximum allowed deviation of pull mean from zero
  fPullWidthThrsh = 2.00;           ///< Maximum allowed deviation of pull width from unity
  fEffThrsh       = 0.50;           ///< Threshold for hit efficiency in the selected range
  fMaxDiffZStHit  = 1.00;           ///< Maximum allowed difference between z-position of hit and station [cm]
  fMinMomentum    = 0.05;           ///< Minimum momentum of particle [GeV/c]
  // Track selection criteria
  fTrackSelectionPrimary  = true;  ///< must the track come from the primary vertex
  fTrackSelectionMomentum = 0.1;   ///< track momentum GeV when it exits the tracking station
  fTrackSelectionAngle    = 60.;   ///< track incl. angle [grad] when it exits the station
}

// ---------------------------------------------------------------------------------------------------------------------
//
void CbmCaInputQaTrd::FillHistogramsPerHit() {}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmCaInputQaTrd::InitDataBranches()
{
  // STS tracking detector interface
  fpDetInterface = CbmTrdTrackingInterface::Instance();

  auto baseInitStatus = CbmCaInputQaBase::InitDataBranches();
  if (kSUCCESS != baseInitStatus) { return baseInitStatus; }

  return kSUCCESS;
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmCaInputQaTrd::InitCanvases()
{
  auto baseInitStatus = CbmCaInputQaBase::InitCanvases();
  if (kSUCCESS != baseInitStatus) { return baseInitStatus; }

  return kSUCCESS;
}

// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmCaInputQaTrd::InitHistograms()
{
  auto baseInitStatus = CbmCaInputQaBase::InitHistograms();
  if (kSUCCESS != baseInitStatus) { return baseInitStatus; }

  return kSUCCESS;
}
