/* Copyright (C) 2023-2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: S.Gorbunov[committer], N.Bluhme */

/// @file    CbmBbaAlignmentTask.cxx
/// @author  Sergey Gorbunov
/// @author  Nora Bluhme
/// @date    20.01.2023
/// @brief   Task class for alignment
///

// Cbm Headers ----------------------
#include "CbmBbaAlignmentTask.h"


// ROOT headers

#include "FairRootManager.h"

#include "TClonesArray.h"
#include "TFile.h"
#include "TH1F.h"
#include "TRandom.h"

//
#include "CbmCaTimeSliceReader.h"
#include "CbmGlobalTrack.h"
#include "CbmL1.h"
#include "CbmMuchTrack.h"
#include "CbmMuchTrackingInterface.h"
#include "CbmMvdHit.h"
#include "CbmMvdTrackingInterface.h"
#include "CbmStsHit.h"
#include "CbmStsSetup.h"
#include "CbmStsTrack.h"
#include "CbmStsTrackingInterface.h"
#include "CbmTofTrack.h"
#include "CbmTofTrackingInterface.h"
#include "CbmTrdTrack.h"
#include "CbmTrdTrackingInterface.h"

#include <cmath>

#include "bba/BBA.h"

// c++ and std headers

#include <iostream>

namespace
{
  using namespace cbm::algo;
}

void CbmBbaAlignmentTask::TrackContainer::MakeConsistent()
{
  fNmvdHits  = 0;
  fNstsHits  = 0;
  fNmuchHits = 0;
  fNtrdHits  = 0;
  fNtofHits  = 0;
  for (const auto& n : fUnalignedTrack.fNodes) {
    switch (n.fSystemId) {
      case ECbmModuleId::kMvd: fNmvdHits++; break;
      case ECbmModuleId::kSts: fNstsHits++; break;
      case ECbmModuleId::kMuch: fNmuchHits++; break;
      case ECbmModuleId::kTrd: fNtrdHits++; break;
      case ECbmModuleId::kTof: fNtofHits++; break;
      default: break;
    }
  }
}


CbmBbaAlignmentTask::CbmBbaAlignmentTask(const char* name, Int_t iVerbose, TString histoFileName)
  : FairTask(name, iVerbose)
  , fHistoFileName(histoFileName)
{
  TFile* curFile           = gFile;
  TDirectory* curDirectory = gDirectory;

  if (!(fHistoFileName == "")) { fHistoFile = new TFile(fHistoFileName.Data(), "RECREATE"); }
  else {
    fHistoFile = gFile;
  }

  fHistoFile->cd();

  fHistoDir = fHistoFile->mkdir("Bba");
  fHistoDir->cd();

  hTestHisto = new TH1F("hTestHisto", "hTestHisto", 100, 0., 1.);

  gFile      = curFile;
  gDirectory = curDirectory;
}


CbmBbaAlignmentTask::~CbmBbaAlignmentTask() {}


InitStatus CbmBbaAlignmentTask::Init()
{

  fFitter.Init();

  //Get ROOT Manager
  FairRootManager* ioman = FairRootManager::Instance();

  if (!ioman) {
    LOG(error) << "CbmBbaAlignmentTask::Init :: RootManager not instantiated!";
    return kERROR;
  }

  // Get global tracks & matches

  if (TrackingMode::kSts != fTrackingMode) {

    fInputGlobalTracks = static_cast<TClonesArray*>(ioman->GetObject("GlobalTrack"));
    if (!fInputGlobalTracks) {
      LOG(error) << "CbmBbaAlignmentTask::Init: Global track array not found!";
      return kERROR;
    }

    fInputGlobalTrackMatches = static_cast<TClonesArray*>(ioman->GetObject("GlobalTrackMatch"));

    if (!fInputGlobalTrackMatches) {
      LOG(error) << "CbmBbaAlignmentTask::Init: Global track matches array not found!";
      //return kERROR;
    }
  }

  fInputStsTracks = static_cast<TClonesArray*>(ioman->GetObject("StsTrack"));
  if (!fInputStsTracks) {
    LOG(error) << "CbmBbaAlignmentTask::Init: Sts track array not found!";
    return kERROR;
  }

  fInputStsTrackMatches = static_cast<TClonesArray*>(ioman->GetObject("StsTrackMatch"));

  if (!fInputStsTrackMatches) {
    LOG(error) << "CbmBbaAlignmentTask::Init: Sts track matches array not found!";
    //return kERROR;
  }

  // MC tracks
  fInputMcTracks = static_cast<TClonesArray*>(ioman->GetObject("MCTrack"));
  if (!fInputMcTracks) {
    Warning("CbmBbaAlignmentTask::Init", "mc track array not found!");
    return kERROR;
  }

  fTracks.clear();

  return kSUCCESS;
}

void CbmBbaAlignmentTask::Exec(Option_t* /*opt*/)
{

  LOG(info) << "BBA: exec event N " << fNEvents;

  fNEvents++;

  if (static_cast<int>(fTracks.size()) >= fMaxNtracks) { return; }

  ca::Framework* l1           = CbmL1::Instance()->fpAlgo;
  const ca::Parameters& l1Par = l1->GetParameters();

  static int statGlobalTracks = 0;
  statGlobalTracks += fInputGlobalTracks->GetEntriesFast();

  // select tracks for alignment and store them
  if (fTrackingMode == kSts && fInputStsTracks) {
    for (int iTr = 0; iTr < fInputStsTracks->GetEntriesFast(); iTr++) {
      if (static_cast<int>(fTracks.size()) >= fMaxNtracks) { break; }
      const CbmStsTrack* stsTrack = dynamic_cast<CbmStsTrack*>(fInputStsTracks->At(iTr));
      if (!stsTrack) continue;
      TrackContainer t;
      if (!fFitter.CreateMvdStsTrack(t.fUnalignedTrack, *stsTrack)) continue;
      t.MakeConsistent();
      fFitter.FitTrack(t.fUnalignedTrack);
      t.fAlignedTrack = t.fUnalignedTrack;
      if (t.fNstsHits < l1Par.GetNstationsActive(ca::EDetectorID::kSts)) continue;
      const auto& par = t.fUnalignedTrack.fNodes[t.fUnalignedTrack.fFirstHitNode].fTrack;
      if (!std::isfinite((ca::fscal) par.GetQp()[0])) continue;
      if (fabs(par.GetQp()[0]) > 1.) continue;  // select tracks with min 1 GeV momentum
      fTracks.push_back(t);
    }
  }
  else if (fTrackingMode == kMcbm && fInputGlobalTracks) {
    for (int iTr = 0; iTr < fInputGlobalTracks->GetEntriesFast(); iTr++) {
      if (static_cast<int>(fTracks.size()) >= fMaxNtracks) { break; }
      const CbmGlobalTrack* globalTrack = dynamic_cast<const CbmGlobalTrack*>(fInputGlobalTracks->At(iTr));
      if (!globalTrack) {
        LOG(fatal) << "BBA: null pointer to the global track!";
        break;
      }
      TrackContainer t;
      if (!fFitter.CreateGlobalTrack(t.fUnalignedTrack, *globalTrack)) {
        LOG(fatal) << "BBA: can not create the global track for the fit! ";
        break;
      }
      t.MakeConsistent();
      t.fUnalignedTrack.fMsQp0                                              = 1. / 0.1;
      t.fUnalignedTrack.fIsMsQp0Set                                         = true;
      t.fUnalignedTrack.fNodes[t.fUnalignedTrack.fFirstHitNode].fTrack.Qp() = 0.;

      fFitter.FitTrack(t.fUnalignedTrack);
      t.fAlignedTrack = t.fUnalignedTrack;
      //if (t.fNstsHits < 1) continue;
      //if (t.fNtrdHits < 2) continue;
      fTracks.push_back(t);
    }
  }

  LOG(info) << "Bba: " << fInputGlobalTracks->GetEntriesFast() << " global tracks in event, " << statGlobalTracks
            << " global tracks in total, " << fTracks.size() << " tracks selected for alignment";
}


void CbmBbaAlignmentTask::ApplyAlignment(const std::vector<double>& par)
{
  // apply alignment parameters to the hits

  for (TrackContainer& t : fTracks) {

    for (unsigned int in = 0; in < t.fUnalignedTrack.fNodes.size(); in++) {
      CbmKfTrackFitter::FitNode& node        = t.fUnalignedTrack.fNodes[in];
      CbmKfTrackFitter::FitNode& nodeAligned = t.fAlignedTrack.fNodes[in];
      int iAlignmentModule                   = node.fReference1;
      if (iAlignmentModule < 0) continue;
      assert(nodeAligned.fReference1 == iAlignmentModule);

      double dx = par[3 * iAlignmentModule + 0];
      double dy = par[3 * iAlignmentModule + 1];
      double dz = par[3 * iAlignmentModule + 2];

      nodeAligned.fMxy.SetX(node.fMxy.X() + ca::fvec(dx));
      nodeAligned.fMxy.SetY(node.fMxy.Y() + ca::fvec(dy));
      nodeAligned.fZ = node.fZ + dz;
    }
  }
}


double CbmBbaAlignmentTask::CostFunction(const std::vector<double>& par)
{
  // apply new parameters to the hits

  for (auto& t : fTracks) {
    t.fAlignedTrack = t.fUnalignedTrack;
  }

  ApplyAlignment(par);

  fChi2Total = 0.;
  fNdfTotal  = 0;

  for (auto& t : fTracks) {
    if (!t.fIsActive) continue;
    fFitter.FitTrack(t.fAlignedTrack);
    double chi2 = t.fAlignedTrack.fNodes[t.fAlignedTrack.fFirstHitNode].fTrack.GetChiSq()[0];
    double ndf  = t.fAlignedTrack.fNodes[t.fAlignedTrack.fFirstHitNode].fTrack.GetNdf()[0];
    if (!std::isfinite(chi2) || chi2 <= 0. || ndf <= 0.) {
      chi2 = 0.;
      ndf  = 0.;
    }
    fChi2Total += chi2;
    fNdfTotal += (int) ndf;
  }

  double cost = fChi2Total / (fNdfTotal + 1);
  if (fFixedNdf > 0 && fNdfTotal != fFixedNdf) { cost = 1.e30; }
  LOG(info) << "BBA: cost function:  ndf " << fNdfTotal << ", cost " << cost
            << ", diff to ideal cost: " << cost - fCostIdeal;

  return cost;
}

void CbmBbaAlignmentTask::Finish()
{
  //
  // perform the alignment
  //

  LOG(info) << "BBA: start the alignment procedure with " << fTracks.size() << " tracks ...";

  ca::Framework* l1 = CbmL1::Instance()->fpAlgo;

  const ca::Parameters& l1Param = l1->GetParameters();

  // init alignmemt module

  int nStations = l1Param.GetNstationsActive();

  for (auto& t : fTracks) {
    for (auto& n : t.fUnalignedTrack.fNodes) {
      n.fReference1 = n.fMaterialLayer;  // material layer is currently equal to the active tracking station
    }
    t.fAlignedTrack = t.fUnalignedTrack;
  }

  int nParameters = 3 * nStations;  // x, y, z

  std::vector<bba::Parameter> par(nParameters);

  for (int ip = 0; ip < nParameters; ip++) {
    bba::Parameter& p = par[ip];
    p.SetActive(0);
    p.SetValue(0.);
    p.SetBoundaryMin(-2.);  // +-1 cm alignment range
    p.SetBoundaryMax(2.);
    p.SetStepMin(1.e-6);
    p.SetStepMax(0.1);
    p.SetStep(100.e-4);
  }

  par[0].SetActive(0);  // fix the first station
  par[1].SetActive(0);
  par[2].SetActive(0);

  par[3 * (nStations - 1) + 0].SetActive(0);  // fix the last station
  par[3 * (nStations - 1) + 1].SetActive(0);
  par[3 * (nStations - 1) + 2].SetActive(0);

  // set active parameters for other stations

  for (int i = 1; i < nStations - 1; i++) {
    par[3 * i + 0].SetActive(0);
    par[3 * i + 1].SetActive(0);
    par[3 * i + 2].SetActive(0);
  }

  for (int i = 1; i < nStations - 1; i++) {
    par[3 * i + 0].SetActive(1);
    par[3 * i + 1].SetActive(1);
    par[3 * i + 2].SetActive(1);
  }

  gRandom->SetSeed(1);

  for (int is = 0; is < nStations; is++) {
    bba::Parameter& px = par[3 * is + 0];
    bba::Parameter& py = par[3 * is + 1];
    bba::Parameter& pz = par[3 * is + 2];

    // +- 0.5 cm misalignment
    double d = 10.;
    if (px.IsActive()) { px.SetValue(gRandom->Uniform(2. * d) - d); }
    if (py.IsActive()) { py.SetValue(gRandom->Uniform(2. * d) - d); }
    if (pz.IsActive()) { pz.SetValue(gRandom->Uniform(2. * d) - d); }
  }

  bba::BBA alignment;

  alignment.setParameters(par);

  auto lambdaCostFunction = [this](const std::vector<double>& p) { return CostFunction(p); };

  alignment.setChi2(lambdaCostFunction);
  alignment.printInfo();

  std::vector<double> parAligned;
  std::vector<double> parMisaligned;
  {
    const std::vector<double>& r = alignment.getResult();
    for (unsigned int i = 0; i < r.size(); i++) {
      parMisaligned.push_back(r[i]);
      parAligned.push_back(0.);
    }
  }

  {
    std::cout << "initial cost function..." << std::endl;
    fCostInitial = CostFunction(parMisaligned);
    for (auto& t : fTracks) {
      double ndf  = t.fAlignedTrack.fNodes[t.fAlignedTrack.fFirstHitNode].fTrack.GetNdf()[0];
      double chi2 = t.fAlignedTrack.fNodes[t.fAlignedTrack.fFirstHitNode].fTrack.GetChiSq()[0];
      if (ndf < 0. || chi2 < 0. || !std::isfinite(ndf) || !std::isfinite(chi2)) { t.fIsActive = false; }
    }
    std::cout << "recalculate initial cost function..." << std::endl;
    fCostInitial = CostFunction(parMisaligned);
    fFixedNdf    = -1;  //fNdfTotal;
  }

  std::cout << "ideal cost function..." << std::endl;
  fCostIdeal = CostFunction(parAligned);


  LOG(info) << "Initial nTracks: " << fTracks.size() << " cost " << fCostInitial;
  LOG(info) << " cost function for the true parameters: " << fCostIdeal;

  alignment.align();

  LOG(info) << " cost function for the true parameters: " << fCostIdeal;

  LOG(info) << " Misaligned parameters: ";
  for (int is = 0; is < nStations; is++) {
    const std::vector<double>& r = parMisaligned;
    LOG(info) << "Sts station " << is << ": x " << r[3 * is + 0] << " y " << r[3 * is + 1] << " z " << r[3 * is + 2];
  }

  LOG(info) << " Alignment results: ";

  for (int is = 0; is < nStations; is++) {
    const std::vector<double>& r = alignment.getResult();
    LOG(info) << "Sts station " << is << ": x " << r[3 * is + 0] << " y " << r[3 * is + 1] << " z " << r[3 * is + 2];
  }

  // store the histograms

  TDirectory* curr   = gDirectory;
  TFile* currentFile = gFile;
  // Open output file and write histograms

  fHistoFile->cd();
  WriteHistosCurFile(fHistoDir);
  if (!(fHistoFileName == "")) {
    fHistoFile->Close();
    fHistoFile->Delete();
  }
  gFile      = currentFile;
  gDirectory = curr;
}

void CbmBbaAlignmentTask::WriteHistosCurFile(TObject* obj)
{
  if (!obj->IsFolder()) obj->Write();
  else {
    TDirectory* cur    = gDirectory;
    TFile* currentFile = gFile;

    TDirectory* sub = cur->GetDirectory(obj->GetName());
    sub->cd();
    TList* listSub = (static_cast<TDirectory*>(obj))->GetList();
    TIter it(listSub);
    while (TObject* obj1 = it()) {
      WriteHistosCurFile(obj1);
    }
    cur->cd();
    gFile      = currentFile;
    gDirectory = cur;
  }
}

ClassImp(CbmBbaAlignmentTask);
