/* Copyright (C) 2023-2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: S.Gorbunov[committer], N.Bluhme */

/// @file    CbmBbaAlignmentTask.h
/// @author  Sergey Gorbunov
/// @author  Nora Bluhme
/// @date    20.01.2023
/// @brief   Task class for alignment
///

#ifndef CbmBbaAlignmentTask_H
#define CbmBbaAlignmentTask_H


#include "CbmKfTrackFitter.h"

#include "FairTask.h"

#include "TString.h"

#include <vector>

class TObject;
class TH1F;
class TClonesArray;
class TFile;
class TDirectory;
class TH1F;


///
/// an example of alignment using BBA package
///
/// you need to switch to the double precision in /algo/ca/CaSimdVc.h
/// by uncommenting this line there:
///
/// typedef Vc::double_v fvec;
///
class CbmBbaAlignmentTask : public FairTask {
public:
  // Constructors/Destructors ---------
  CbmBbaAlignmentTask(const char* name = "CbmBbaAlignmentTask", Int_t iVerbose = 0,
                      TString histoFileName = "CbmBbaAlignmentHisto.root");
  ~CbmBbaAlignmentTask();

  Int_t GetZtoNStation(Double_t getZ);

  InitStatus Init();
  void Exec(Option_t* opt);
  void Finish();

  void SetMcbmTrackingMode() { fTrackingMode = TrackingMode::kMcbm; }
  void SetStsTrackingMode() { fTrackingMode = TrackingMode::kSts; }

public:
  enum TrackingMode
  {
    kSts,
    kMcbm
  };

  /// information about a track
  /// aligned and unaligned hits are stored in the corresponding CbmKfTrackFitter::Track objects
  struct TrackContainer {

    CbmKfTrackFitter::Track fUnalignedTrack;  // track before alignment
    CbmKfTrackFitter::Track fAlignedTrack;    // track after alignment

    int fNmvdHits {0};   // number of MVD hits
    int fNstsHits {0};   // number of STS hits
    int fNmuchHits {0};  // number of MUCH hits
    int fNtrdHits {0};   // number of TRD hits
    int fNtofHits {0};   // number of TOF hits

    bool fIsActive {1};  // is the track active
    void MakeConsistent();
  };

private:
  const CbmBbaAlignmentTask& operator=(const CbmBbaAlignmentTask&);
  CbmBbaAlignmentTask(const CbmBbaAlignmentTask&);

  void WriteHistosCurFile(TObject* obj);

  void ApplyAlignment(const std::vector<double>& par);

  double CostFunction(const std::vector<double>& par);

  TrackingMode fTrackingMode = TrackingMode::kMcbm;

  // input data arrays

  TClonesArray* fInputGlobalTracks {nullptr};
  TClonesArray* fInputStsTracks {nullptr};

  TClonesArray* fInputMcTracks {nullptr};            // Mc info for debugging
  TClonesArray* fInputGlobalTrackMatches {nullptr};  // Mc info for debugging
  TClonesArray* fInputStsTrackMatches {nullptr};     // Mc info for debugging

  CbmKfTrackFitter fFitter;

  // collection of selected tracks and hits
  std::vector<TrackContainer> fTracks;

  //output file with histograms
  TString fHistoFileName {"CbmBbaAlignmentHisto.root"};
  TFile* fHistoFile {nullptr};
  TDirectory* fHistoDir {nullptr};

  Int_t fNEvents {0};

  Int_t fMaxNtracks {10000};

  double fCostIdeal {1.e10};
  double fCostInitial {0.};

  double fChi2Total {0.};
  long fNdfTotal {0};
  long fFixedNdf {-1};

  //histograms
  TH1F* hTestHisto {nullptr};

  ClassDef(CbmBbaAlignmentTask, 1);
};

#endif
