/* Copyright (C) 2023 Facility for Antiproton and Ion Research in Europe, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Pierre-Alain Loizeau [committer] */

#include "Application.h"

#include "CbmFlesCanvasTools.h"

#include <Logger.h>

#include "TCanvas.h"
#include "TEnv.h"
#include "TFile.h"
#include "TH1.h"
#include "TH2.h"
#include "THttpServer.h"
#include "TMessage.h"
#include "TObjArray.h"
#include "TProfile.h"
#include "TRootSniffer.h"
#include "TSystem.h"

#include <boost/archive/binary_iarchive.hpp>
#include <boost/iostreams/device/array.hpp>
#include <boost/iostreams/stream.hpp>
#include <boost/serialization/utility.hpp>
#include <boost/serialization/vector.hpp>

#include <mutex>
#include <zmq_addon.hpp>

#include "Histo1D.h"
#include "ui_callbacks.h"

std::mutex mtx;

namespace b_io = boost::iostreams;
namespace b_ar = boost::archive;

namespace cbm::services::histserv
{
  // -----   Constructor   ---------------------------------------------------------------------------------------------
  Application::Application(ProgramOptions const& opt) : fOpt(opt)
  {
    /// Read options from executable
    LOG(info) << "Options for Application:";
    LOG(info) << " Input ZMQ channel: " << fOpt.ComChan();
    LOG(info) << " HTTP server port:  " << fOpt.HttpPort();
    if ("" != fOpt.HistoFile()) {  //
      LOG(info) << " Output filename:   " << fOpt.HistoFile() << (fOpt.Overwrite() ? " (in overwrite mode)" : "");
    }

    /// FIXME: SOMETHING_To_Replace_FairMQ!!!!!!!!!!!!!
    /// FIXME: Initialize communication channels of SOMETHING_To_Replace_FairMQ
    /// FIXME: Link channel to method in order to process received messages
    // fZmqSocket.set(zmq::sockopt::rcvhwm, int(hwm));  // FIXME: need for HWM?
    fZmqSocket.set(zmq::sockopt::rcvtimeo, fOpt.ComChanZmqRcvTo());  // Timeout in ms to avoid stuck in loop!
    fZmqSocket.bind(fOpt.ComChan().c_str());  // This side "binds" the socket => Other side should connect!!!!

    fServer = new THttpServer(Form("http:%u", fOpt.HttpPort()));
    /// To avoid the server sucking all Histos from gROOT when no output file is used
    fServer->GetSniffer()->SetScanGlobalDir(kFALSE);
    const char* jsrootsys = gSystem->Getenv("JSROOTSYS");
    if (!jsrootsys) jsrootsys = gEnv->GetValue("HttpServ.JSRootPath", jsrootsys);

    fUiCmdActor = new UiCmdActor();
    fServer->Register("/", fUiCmdActor);
    fServer->Hide("/UiCmdActor");

    fServer->RegisterCommand("/Reset_Hist", "/UiCmdActor/->SetResetHistos()");
    fServer->RegisterCommand("/Save_Hist", "/UiCmdActor/->SetSaveHistos()");
    fServer->RegisterCommand("/Stop_Server", "/UiCmdActor/->SetServerStop()");

    /*
    fServer->RegisterCommand("/Reset_Hist", "this->ResetHistograms()");
    fServer->RegisterCommand("/Save_Hist", "this->SaveHistograms()");
    */

    fServer->Restrict("/Reset_Hist", "allow=admin");
    fServer->Restrict("/Save_Hist", "allow=admin");
    fServer->Restrict("/Stop_Server", "allow=admin");


    LOG(info) << "JSROOT location: " << jsrootsys;
  }
  // -------------------------------------------------------------------------------------------------------------------


  // -----   Main Loop   -----------------------------------------------------------------------------------------------
  void Application::Exec()
  {
    fStopThread = false;
    fThread     = std::thread(&Application::UpdateHttpServer, this);
    while (!(fUiCmdActor->GetServerStop())) {  //
      /// Infinite loop, this is a service which should survive until told otherwise after all

      /// FIXME: Start listening to <SOMETHING?!?> to receive histograms and configuration
      /// FIXME: handle signals from OS/console
      /* Jan suggestion with zmq_addon CPP interface */
      std::vector<zmq::message_t> vMsg;
      const auto ret = zmq::recv_multipart(fZmqSocket, std::back_inserter(vMsg));
      if (!ret) continue;

      std::lock_guard<std::mutex> lk(mtx);
      if (*ret > 3) {  //
        ReceiveConfigAndData(vMsg);
      }
      else if (*ret == 1) {  //
        ReceiveData(vMsg[0]);
      }
      else {  //
        LOG(error) << "Invalid number of message parts received: should be either 1 or more than 3 vs " << *ret;
      }
    }
  }
  // -------------------------------------------------------------------------------------------------------------------

  // -----   Constructor   ---------------------------------------------------------------------------------------------
  Application::~Application()
  {
    SaveHistograms();
    fStopThread = true;
    fThread.join();
    SaveHistograms();
  }
  // -------------------------------------------------------------------------------------------------------------------

  // -----   Server update background thread   -------------------------------------------------------------------------
  void Application::UpdateHttpServer()
  {
    /// This is needed to have a reactive GUI independently of histogram updates reception
    while (!fStopThread) {
      std::this_thread::sleep_for(std::chrono::milliseconds(10));
      std::lock_guard<std::mutex> lk(mtx);

      fServer->ProcessRequests();

      /// TODO: control flags communication from histo server to histograms sources?
      /// Idea: 1 req channel (per process or not, mixup?), polling every N TS and/or M s
      if (fUiCmdActor->GetResetHistos()) {
        LOG(info) << "Reset Monitor histos ";
        ResetHistograms();
        fUiCmdActor->SetResetHistos(false);
      }  // if( fUiCmdActor->GetResetHistos() )

      if (fUiCmdActor->GetSaveHistos()) {
        LOG(info) << "Save All histos & canvases";
        SaveHistograms();
        fUiCmdActor->SetSaveHistos(false);
      }  // if( fUiCmdActor->GetSaveHistos() )
    }
  }
  // -------------------------------------------------------------------------------------------------------------------

  // -----   Server update background thread   -------------------------------------------------------------------------
  bool Application::ReceiveData(zmq::message_t& msg)
  {
    LOG(debug) << "Application::ReceiveData => Processing histograms update";

    /// FIXME: Something to replace FairMQ and extract the histograms!!!!
    /// FIXME: Need something to replace the ROOT serializer which allowed to have any of TH1x, TH2x, TH3x or TProfile
    /// FIXME: Need something to replace the TObjArray which allowed to have a mix of of TH1x, TH2x, TH3x or TProfile
    b_io::basic_array_source<char> device(static_cast<char*>(msg.data()), msg.size());
    b_io::stream<b_io::basic_array_source<char>> s(device);
    b_ar::binary_iarchive iarch(s);

    std::vector<cbm::algo::Histo1D> vHist;
    iarch >> vHist;

    for (auto& source : vHist) {
      /// copied from CbmTaskDigiEventQa::ToTH1D
      /// FIXME: Should be placed in a tools/interface/whatever library with all similar functions!!
      /// FIXME: Reverse OP need to be implemented + CI unit tests for back and forth in each direction (ROOT <-> Algo)
      /// FIXME: Lead to "Warning in <TROOT::Append>: Replacing existing TH1: xxxxxx (Potential memory leak)."
      bool add = TH1::AddDirectoryStatus();
      TH1::AddDirectory(false);  // Needed to prevent ROOT from adding histogram to its internal registry
      TH1D* result = new TH1D(source.Name().c_str(), source.Title().c_str(),  // Titles in ROOT convention with ";" sep
                              source.NumBins(), source.MinValue(), source.MaxValue());
      TH1::AddDirectory(add);  // Needed to prevent ROOT from adding histogram to its internal registry
      for (uint32_t bin = 0; bin < source.NumBins(); bin++) {
        result->SetBinContent(1 + bin, source.Content(bin));
      }
      result->SetBinContent(0, source.Underflow());
      result->SetBinContent(source.NumBins(), source.Overflow());
      result->SetEntries(source.NumEntries());
      if (!ReadHistogram<TH1>(result)) {  //
        return false;
      }
      delete result;
    }

    /// If new histos received, try to prepare as many canvases as possible
    /// Should be expensive on start and cheap afterward
    if (!fbAllCanvasReady) {
      LOG(debug) << "Application::ReceiveData => Checking for canvases updates";
      for (uint32_t uCanv = 0; uCanv < fvpsCanvasConfig.size(); ++uCanv) {
        /// Jump canvases already ready
        if (fvbCanvasReady[uCanv]) {  //
          continue;
        }

        /// Now come the expensive part as we unpack its config and check each histo
        fvbCanvasReady[uCanv] = PrepareCanvas(uCanv);
      }  // for( uint32_t uCanv = 0; uCanv < fvpsCanvasConfig.size(); ++uCanv )
    }    // if( !fbAllCanvasReady )

    /*
    TObject* tempObject = nullptr;
    if (TString(tempObject->ClassName()).EqualTo("TObjArray")) {
      std::lock_guard<std::mutex> lk(mtx);
      TObjArray* arrayHisto = static_cast<TObjArray*>(tempObject);
      for (Int_t i = 0; i < arrayHisto->GetEntriesFast(); i++) {
        TObject* pObj = arrayHisto->At(i);

        if (nullptr != dynamic_cast<TProfile*>(pObj)) {
          if (!ReadHistogram<TProfile>(dynamic_cast<TProfile*>(pObj))) {  //
            return false;
          }
        }  // if( nullptr != dynamic_cast< TProfile *>( pObj ) )
        else if (nullptr != dynamic_cast<TH2*>(pObj)) {
          if (!ReadHistogram<TH2>(dynamic_cast<TH2*>(pObj))) {  //
            return false;
          }
        }  // if( nullptr != dynamic_cast< TH2 *>( pObj ) )
        else if (nullptr != dynamic_cast<TH1*>(pObj)) {
          if (!ReadHistogram<TH1>(dynamic_cast<TH1*>(pObj))) {  //
            return false;
          }
        }  // if( nullptr != dynamic_cast< TH1 *>( pObj ) )
        else
          LOG(warning) << "Unsupported object type for " << pObj->GetName();
      }  // for (Int_t i = 0; i < arrayHisto->GetEntriesFast(); i++)

      LOG(debug) << "Application::ReceiveData => Deleting array";
      /// Need to use Delete instead of Clear to avoid memory leak!!!
      arrayHisto->Delete();

      /// If new histos received, try to prepare as many canvases as possible
      /// Should be expensive on start and cheap afterward
      if (!fbAllCanvasReady) {
        LOG(debug) << "Application::ReceiveData => Checking for canvases updates";
        for (uint32_t uCanv = 0; uCanv < fvpsCanvasConfig.size(); ++uCanv) {
          /// Jump canvases already ready
          if (fvbCanvasReady[uCanv]) {  //
            continue;
          }

          /// Now come the expensive part as we unpack its config and check each histo
          fvbCanvasReady[uCanv] = PrepareCanvas(uCanv);
        }  // for( uint32_t uCanv = 0; uCanv < fvpsCanvasConfig.size(); ++uCanv )
      }    // if( !fbAllCanvasReady )
    }      // if (TString(tempObject->ClassName()).EqualTo("TObjArray"))
    else {
      fStopThread         = true;
      std::string err_msg = "Application::ReceiveData => Wrong object type at input: ";
      err_msg += tempObject->ClassName();
      throw std::runtime_error(err_msg);
    }

    if (nullptr != tempObject) delete tempObject;
    */

    fNMessages += 1;

    LOG(debug) << "Application::ReceiveData => Finished processing histograms update";

    return true;
  }

  bool Application::ReceiveHistoConfig(zmq::message_t& msg)
  {
    /// FIXME: Something to replace FairMQ and extract the config!!!!
    //  BoostSerializer<std::pair<std::string, std::string>>().Deserialize(msg, tempObject);
    b_io::basic_array_source<char> device(static_cast<char*>(msg.data()), msg.size());
    b_io::stream<b_io::basic_array_source<char>> s(device);
    b_ar::binary_iarchive iarch(s);

    std::pair<std::string, std::string> tempObject("", "");
    iarch >> tempObject;

    LOG(info) << " Received configuration for histo " << tempObject.first << " : " << tempObject.second;

    /// Check if histo name already received in previous messages
    /// Linear search should be ok as config is shared only at startup
    UInt_t uPrevHist = 0;
    for (uPrevHist = 0; uPrevHist < fvpsHistosFolder.size(); ++uPrevHist) {
      if (fvpsHistosFolder[uPrevHist].first == tempObject.first) {  //
        break;
      }
    }  // for( UInt_t uPrevHist = 0; uPrevHist < fvpsHistosFolder.size(); ++uPrevHist )

    if (uPrevHist < fvpsHistosFolder.size()) {
      LOG(info) << " Ignored new configuration for histo " << tempObject.first
                << " due to previously received one: " << tempObject.second;
      /// Not sure if we should return false here...
    }  // if( uPrevHist < fvpsHistosFolder.size() )
    else {
      fvpsHistosFolder.push_back(tempObject);
      fvHistos.push_back(std::pair<TNamed*, std::string>(nullptr, ""));
      fvbHistoRegistered.push_back(false);
      fbAllHistosRegistered = false;
    }  // else of if( uPrevHist < fvpsHistosFolder.size() )

    return true;
  }

  bool Application::ReceiveCanvasConfig(zmq::message_t& msg)
  {
    /// FIXME: Something to replace FairMQ and extract the config!!!!
    //  BoostSerializer<std::pair<std::string, std::string>>().Deserialize(msg, tempObject);
    b_io::basic_array_source<char> device(static_cast<char*>(msg.data()), msg.size());
    b_io::stream<b_io::basic_array_source<char>> s(device);
    b_ar::binary_iarchive iarch(s);

    std::pair<std::string, std::string> tempObject("", "");
    iarch >> tempObject;

    LOG(info) << " Received configuration for canvas " << tempObject.first << " : " << tempObject.second;

    /// Check if canvas name already received in previous messages
    /// Linear search should be ok as config is shared only at startup
    uint32_t uPrevCanv = 0;
    for (uPrevCanv = 0; uPrevCanv < fvpsCanvasConfig.size(); ++uPrevCanv) {
      if (fvpsCanvasConfig[uPrevCanv].first == tempObject.first) {  //
        break;
      }
    }  // for( UInt_t uPrevCanv = 0; uPrevCanv < fvpsCanvasConfig.size(); ++uPrevCanv )

    if (uPrevCanv < fvpsCanvasConfig.size()) {
      LOG(warning) << " Ignored new configuration for Canvas " << tempObject.first
                   << " due to previously received one: " << tempObject.second;
      /// Not sure if we should return false here...
    }  // if( uPrevCanv < fvpsCanvasConfig.size() )
    else {
      fvpsCanvasConfig.push_back(tempObject);
      fvbCanvasReady.push_back(false);
      fbAllCanvasReady = false;

      fvCanvas.push_back(std::pair<TCanvas*, std::string>(nullptr, ""));
      fvbCanvasRegistered.push_back(false);
      fbAllCanvasRegistered = false;
    }  // else of if( uPrevCanv < fvpsCanvasConfig.size() )
    return true;
  }

  bool Application::ReceiveConfigAndData(std::vector<zmq::message_t>& vMsg)
  {
    /// FIXME: Something to replace FairMQ and extract the histograms!!!!
    LOG(debug) << "Application::ReceiveConfigAndData => Received composed message with " << vMsg.size() << " parts";

    /// Header contains a pair of unsigned integers
    /// FIXME: Something to replace FairMQ and extract the header!!!!
    // BoostSerializer<std::pair<uint32_t, uint32_t>>().Deserialize(vMsg.at(0), pairHeader);
    b_io::basic_array_source<char> device_header(static_cast<char*>(vMsg.at(0).data()), vMsg.at(0).size());
    b_io::stream<b_io::basic_array_source<char>> s_header(device_header);
    b_ar::binary_iarchive iarch_header(s_header);

    std::pair<uint32_t, uint32_t> pairHeader;
    iarch_header >> pairHeader;
    LOG(debug) << "Application::ReceiveConfigAndData => Received configuration for " << pairHeader.first
               << " histos and " << pairHeader.second << " canvases";

    uint32_t uOffsetHistoConfig = pairHeader.first;
    if (0 == pairHeader.first) {
      uOffsetHistoConfig = 1;
      if (0 < vMsg[uOffsetHistoConfig].size()) {
        fStopThread         = true;
        fUiCmdActor->SetServerStop();
        std::string err_msg = "Application::ReceiveConfigAndData => No histo config expected but corresponding message";
        err_msg += " is not empty: ";
        err_msg += vMsg[uOffsetHistoConfig].size();
        throw std::runtime_error(err_msg);
      }
    }

    uint32_t uOffsetCanvasConfig = pairHeader.second;
    if (0 == pairHeader.second) {
      uOffsetCanvasConfig = 1;
      if (0 < vMsg[uOffsetHistoConfig + uOffsetCanvasConfig].size()) {
        fStopThread         = true;
        fUiCmdActor->SetServerStop();
        std::string err_msg = "Application::ReceiveConfigAndData => No Canvas config expected but corresponding ";
        err_msg += " message is not empty: ";
        err_msg += vMsg[uOffsetHistoConfig + uOffsetCanvasConfig].size();
        throw std::runtime_error(err_msg);
      }
    }

    if ((1 + uOffsetHistoConfig + uOffsetCanvasConfig + 1) != vMsg.size()) {
      fStopThread         = true;
      fUiCmdActor->SetServerStop();
      std::string err_msg = "Application::ReceiveConfigAndData => Nb parts in message not matching configs numbers ";
      err_msg += " declared in header";
      err_msg += vMsg.size();
      err_msg += " VS ";
      err_msg += 1 + uOffsetHistoConfig + uOffsetCanvasConfig + 1;
      throw std::runtime_error(err_msg);
    }

    /// Decode parts for histograms configuration (auto-skip empty message if 0 declared in header)
    for (uint32_t uHisto = 0; uHisto < pairHeader.first; ++uHisto) {
      ReceiveHistoConfig(vMsg[1 + uHisto]);
    }  // for (UInt_t uHisto = 0; uHisto < pairHeader.first; ++uHisto)
    LOG(debug) << "Application::ReceiveConfigAndData => Processed configuration for " << pairHeader.first << " histos";

    /// Decode parts for histograms configuration (auto-skip empty message if 0 declared in header)
    for (uint32_t uCanv = 0; uCanv < pairHeader.second; ++uCanv) {
      ReceiveCanvasConfig(vMsg[1 + uOffsetHistoConfig + uCanv]);
    }  // for (UInt_t uCanv = 0; uCanv < pairHeader.second; ++uCanv)
    LOG(debug) << "Application::ReceiveConfigAndData => Processed configuration for " << pairHeader.second
               << " canvases";

    /// Decode the histograms data now that the configuration is loaded
    ReceiveData(vMsg[1 + uOffsetHistoConfig + uOffsetCanvasConfig]);

    return true;
  }
  // -------------------------------------------------------------------------------------------------------------------

  template<class HistoT>
  bool Application::ReadHistogram(HistoT* pHist)
  {
    int index1 = FindHistogram(pHist->GetName());
    if (-1 == index1) {
      HistoT* histogram_new = static_cast<HistoT*>(pHist->Clone());
      fArrayHisto.Add(histogram_new);

      LOG(info) << "Received new histo " << pHist->GetName();

      /// If new histo received, try to register it if configuration available
      if (!fbAllHistosRegistered) {
        for (uint32_t uHist = 0; uHist < fvpsHistosFolder.size(); ++uHist) {
          /// Jump histos already ready
          if (fvbHistoRegistered[uHist]) {  //
            continue;
          }

          /// Check if name matches one in config for others
          if (fvpsHistosFolder[uHist].first == histogram_new->GetName()) {
            fvHistos[uHist] = std::pair<TNamed*, std::string>(histogram_new, fvpsHistosFolder[uHist].second);
            fServer->Register(Form("/%s", fvHistos[uHist].second.data()), fvHistos[uHist].first);
            fvbHistoRegistered[uHist] = true;

            LOG(info) << "registered histo " << fvHistos[uHist].first->GetName() << " in folder "
                      << fvHistos[uHist].second;


            /// Update flag telling whether all known histos are registered
            fbAllHistosRegistered = true;
            for (uint32_t uIdx = 0; uIdx < fvbHistoRegistered.size(); ++uIdx) {
              if (!fvbHistoRegistered[uIdx]) {
                fbAllHistosRegistered = false;
                break;
              }  // if( !fvbHistoRegistered[ uIdx ] )
            }    // for( uint32_t uIdx = 0; uIdx < fvbHistoRegistered.size(); ++uIdx )

            break;
          }  // if( fvpsHistosFolder[ uHist ].first == histogram_new->GetName() )
        }    // for( uint32_t uCanv = 0; uCanv < fvpsCanvasConfig.size(); ++uCanv )
      }      // if( !fbAllCanvasReady )
    }        // if (-1 == index1)
    else {
      LOG(debug) << "Received update for: " << pHist->GetName();
      HistoT* histogram_existing = dynamic_cast<HistoT*>(fArrayHisto.At(index1));
      if (nullptr == histogram_existing) {
        LOG(error) << "CbmMqHistoServer::ReadHistogram => "
                   << "Incompatible type found during update for histo " << pHist->GetName();
        return false;
      }  // if( nullptr == histogram_existing )

      histogram_existing->Add(pHist);
    }  // else of if (-1 == index1)
    return true;
  }

  int Application::FindHistogram(const std::string& name)
  {
    for (int iHist = 0; iHist < fArrayHisto.GetEntriesFast(); ++iHist) {
      TObject* obj = fArrayHisto.At(iHist);
      if (TString(obj->GetName()).EqualTo(name)) {  //
        return iHist;
      }
    }  // for( int iHist = 0; iHist < fArrayHisto.GetEntriesFast(); ++iHist )
    return -1;
  }

  bool Application::PrepareCanvas(uint32_t uCanvIdx)
  {
    LOG(info) << " Extracting configuration for canvas index " << uCanvIdx;
    CanvasConfig conf(ExtractCanvasConfigFromString(fvpsCanvasConfig[uCanvIdx].second));

    /// First check if all objects to be drawn are present
    uint32_t uNbPads(conf.GetNbPads());
    for (uint32_t uPadIdx = 0; uPadIdx < uNbPads; ++uPadIdx) {
      uint32_t uNbObj(conf.GetNbObjsInPad(uPadIdx));
      for (uint32_t uObjIdx = 0; uObjIdx < uNbObj; ++uObjIdx) {
        std::string sName(conf.GetObjName(uPadIdx, uObjIdx));
        /// Check for empty pads!
        if ("nullptr" != sName) {
          if (FindHistogram(sName) < 0) {
            return false;
          }  // if( FindHistogram( conf.GetObjName( uPadIdx, uObjIdx ) ) < 0 )
        }    // if( "nullptr" != sName )
      }      // for( uint32_t uObjIdx = 0; uObjIdx < uNbObj; ++uObjIdx )
    }        // for( uint32_t uPadIdx = 0; uPadIdx < uNbPads; ++uPadIdx )

    LOG(info) << " All histos found for canvas " << conf.GetName().data() << ", now preparing it";

    /// Create new canvas and pads
    TCanvas* pNewCanv = new TCanvas(conf.GetName().data(), conf.GetTitle().data());
    pNewCanv->Divide(conf.GetNbPadsX(), conf.GetNbPadsY());

    /// Loop on pads
    for (uint32_t uPadIdx = 0; uPadIdx < uNbPads; ++uPadIdx) {
      pNewCanv->cd(1 + uPadIdx);

      /// Pad settings
      gPad->SetGrid(conf.GetGridx(uPadIdx), conf.GetGridy(uPadIdx));
      gPad->SetLogx(conf.GetLogx(uPadIdx));
      gPad->SetLogy(conf.GetLogy(uPadIdx));
      gPad->SetLogz(conf.GetLogz(uPadIdx));

      /// Add objects (we know they are there
      uint32_t uNbObj(conf.GetNbObjsInPad(uPadIdx));
      for (uint32_t uObjIdx = 0; uObjIdx < uNbObj; ++uObjIdx) {
        std::string sName(conf.GetObjName(uPadIdx, uObjIdx));
        if ("nullptr" != sName) {
          TObject* pObj = fArrayHisto[FindHistogram(sName)];

          if (nullptr != dynamic_cast<TProfile*>(pObj)) {
            dynamic_cast<TProfile*>(pObj)->Draw(conf.GetOption(uPadIdx, uObjIdx).data());
          }  // if( nullptr != dynamic_cast< TProfile *>( pObj ) )
          else if (nullptr != dynamic_cast<TH2*>(pObj)) {
            dynamic_cast<TH2*>(pObj)->Draw(conf.GetOption(uPadIdx, uObjIdx).data());
          }  // if( nullptr != dynamic_cast< TH2 *>( pObj ) )
          else if (nullptr != dynamic_cast<TH1*>(pObj)) {
            dynamic_cast<TH1*>(pObj)->Draw(conf.GetOption(uPadIdx, uObjIdx).data());
          }  // if( nullptr != dynamic_cast< TH1 *>( pObj ) )
          else
            LOG(warning) << "  Unsupported object type for " << sName << " when preparing canvas " << conf.GetName();

          LOG(info) << "  Configured histo " << sName << " on pad " << (1 + uPadIdx) << " for canvas "
                    << conf.GetName().data();
        }  // if( "nullptr" != sName )
      }    // for( uint32_t uObjIdx = 0; uObjIdx < uNbObj; ++uObjIdx )
    }      // for( uint32_t uPadIdx = 0; uPadIdx < uNbPads; ++uPadIdx )

    fvCanvas[uCanvIdx] = std::pair<TCanvas*, std::string>(pNewCanv, "canvases");
    fServer->Register(Form("/%s", fvCanvas[uCanvIdx].second.data()), fvCanvas[uCanvIdx].first);
    fvbCanvasRegistered[uCanvIdx] = true;

    LOG(info) << " Registered canvas " << fvCanvas[uCanvIdx].first->GetName() << " in folder "
              << fvCanvas[uCanvIdx].second;

    /// Update flag telling whether all known canvases are registered
    fbAllCanvasRegistered = true;
    for (uint32_t uIdx = 0; uIdx < fvbCanvasRegistered.size(); ++uIdx) {
      if (!fvbCanvasRegistered[uIdx]) {
        fbAllCanvasRegistered = false;
        break;
      }  // if( !fvbCanvasRegistered[ uIdx ] )
    }    // for( uint32_t uIdx = 0; uIdx < fvbCanvasRegistered.size(); ++uIdx )

    return true;
  }

  bool Application::ResetHistograms()
  {
    for (int iHist = 0; iHist < fArrayHisto.GetEntriesFast(); ++iHist) {
      dynamic_cast<TH1*>(fArrayHisto.At(iHist))->Reset();
    }  // for( int iHist = 0; iHist < fArrayHisto.GetEntriesFast(); ++iHist )
    return true;
  }

  bool Application::SaveHistograms()
  {
    if ("" == fOpt.HistoFile()) {  //
      LOG(error) << "Filename for saving histograms and canvases not defined. Ignoring request.";
      return false;
    }

    /// Save old global file and folder pointer to avoid messing with FairRoot
    TFile* oldFile     = gFile;
    TDirectory* oldDir = gDirectory;

    /// (Re-)Create ROOT file to store the histos
    TFile* histoFile = nullptr;

    // open separate histo file in recreate mode
    histoFile = new TFile(fOpt.HistoFile().data(), fOpt.Overwrite() ? "RECREATE" : "CREATE");

    if (nullptr == histoFile) {  //
      gFile      = oldFile;
      gDirectory = oldDir;
      LOG(error) << "Ignoring request to save histograms and canvases: could not open output file " << fOpt.HistoFile();
      return false;
    }

    LOG(info) << "Saving Histograms and canvases in file: " << fOpt.HistoFile();

    /// Register the histos in the HTTP server
    for (UInt_t uHisto = 0; uHisto < fvHistos.size(); ++uHisto) {
      /// catch case of histograms declared in config but not yet received
      if (nullptr != fvHistos[uHisto].first) {
        /// Make sure we end up in chosen folder
        TString sFolder = fvHistos[uHisto].second.data();
        if (nullptr == gDirectory->Get(sFolder)) {  //
          gDirectory->mkdir(sFolder);
        }
        gDirectory->cd(sFolder);

        /// Write plot
        fvHistos[uHisto].first->Write();
      }

      histoFile->cd();
    }  // for( UInt_t uHisto = 0; uHisto < fvHistos.size(); ++uHisto )

    for (UInt_t uCanvas = 0; uCanvas < fvCanvas.size(); ++uCanvas) {
      /// catch case of canvases declared in config but for which not all histos were yet received
      if (nullptr != fvCanvas[uCanvas].first) {
        /// Make sure we end up in chosen folder
        TString sFolder = fvCanvas[uCanvas].second.data();
        if (nullptr == gDirectory->Get(sFolder)) {  //
          gDirectory->mkdir(sFolder);
        }
        gDirectory->cd(sFolder);

        /// Write plot
        fvCanvas[uCanvas].first->Write();
      }

      histoFile->cd();
    }  // for( UInt_t uHisto = 0; uHisto < fvCanvas.size(); ++uHisto )

    /// Restore old global file and folder pointer to avoid messing with FairRoot
    gFile      = oldFile;
    gDirectory = oldDir;

    histoFile->Close();

    return true;
  }

}  // namespace cbm::services::histserv
